"""
Generates fake endpoints for a specified amount and type of data

The factory can generate dynamic values for endpoints if necessary.
e.g. { "id": <int:10> } would translate into 10 random integers

Possible dynamic values are:
- int
- str
- sha256
- name
- uri
"""

import json
import re

from faker import Faker


class EndpointFactory:

    def __init__(self, raw_data):
        self.raw_data = raw_data
        self.endpoints = []

    def generate(self):
        for details in self.raw_data:
            url = details['url']
            method = details['method']
            response = details['response']

            # TODO: Parse JSON response and see if any dynamic values are present
            # parsed_response = json.loads(response)
            # for k, v in parsed_response:
            #     reg = re.findall(r'<(.*)>', v, flags=re.DOTALL)
            #     endpoint = {}
            #     if reg:
            #         for r in reg:
            #             if ':' in r:
            #                 dynamic_type, instances = r.split(':')
            #                 self._factory_generate(dynamic_type, instances)
            #                 continue
            self.endpoints.append({
                'url': url.strip(),
                'method': method,
                'response': json.dumps(json.loads(response))
            })

    # def _factory_generate(self, dynamic_type, instances):



import os
import subprocess
import sys


def main(index_file):
    # Check if node is installed
    node_exists = subprocess.check_output(['node', '-v'])
    if not node_exists:
        sys.stdout.write('Node installation not found. Please install Node.js')
        sys.exit(0)

    # Check if Express is installed
    try:
        express_exists = subprocess.check_output(['npm', 'list', '-g', 'express'])
    except subprocess.CalledProcessError:
        sys.stdout.write('Express installation not found globally, checking local npm installation...')

        try:
            express_exists = subprocess.check_output(['npm', 'list', 'express'])
        except subprocess.CalledProcessError:
            sys.stdout.write('Express installation not found locally. Please ensure Express is installed')
            sys.exit(0)


    try:
        subprocess.call(["node", index_file])
    except KeyboardInterrupt:
        os.remove(index_file)
        sys.stdout.write('Mock server shutdown.')



"""
Generates mock Express server file
"""

import os

from .extractor import Extractor
from .factory import EndpointFactory


def main(output=None, port=8000):
    extractor = Extractor()
    factory = EndpointFactory(extractor.url_details)
    factory.generate()

    output_path = output if output is not None else 'index.js'

    with open(output_path, 'w') as f:
        f.write('const express = require("express");\n')
        f.write('const app = express();\n')

        for endpoint in factory.endpoints:
            f.write('app.%s("%s", (req, res) => {res.send(%s);});\n' % (
                endpoint['method'],
                endpoint['url'],
                endpoint['response']
            ))

        f.write('app.listen({port}, () => console.log("Django Mock JSON Server: Listening on port {port}"));\n'.format(port=port))



"""
Extracts all endpoints and their corresponding view docstrings to retrieve
the expected JSON responses
"""

import re

import django
from django.conf import settings
from django.core.exceptions import ViewDoesNotExist


if django.VERSION >= (2, 0):
    from django.urls import URLPattern, URLResolver

    class RegexURLPattern:
        pass

    class RegexURLResolver:
        pass

    class LocaleRegexURLResolver:
        pass

    def describe_pattern(p):
        return str(p.pattern)
else:
    try:
        from django.urls import RegexURLPattern, RegexURLResolver, LocaleRegexURLResolver
    except ImportError:
        from django.core.urlresolvers import RegexURLPattern, RegexURLResolver, LocaleRegexURLResolver

    class URLPattern:
        pass

    class URLResolver:
        pass

    def describe_pattern(p):
        return p.regex.pattern


METHODS = ['get', 'post', 'put', 'patch', 'delete']


class Extractor:

    def __init__(self):
        self._urlconf = __import__(getattr(settings, 'ROOT_URLCONF'))
        self.url_details = []
        self._load_url_details()  # Load all urlpatterns & corresponding view classes/functions

    def _get_view_details(self, urlpatterns, parent=''):
        """Recursive function to extract all url details"""
        for pattern in urlpatterns:
            if isinstance(pattern, (URLPattern, RegexURLPattern)):
                try:
                    d = describe_pattern(pattern)
                    full_url = parent + d
                    docstr = pattern.callback.__doc__
                    method = None
                    expected_json_response = ''

                    if not docstr:
                        continue

                    # Get expected URL
                    u = re.findall(r'URL: (.*)', docstr, flags=re.DOTALL)
                    if u:
                        expected_url = u[0]
                    else:
                        expected_url = ''

                    # Get all possible methods
                    if 'view_class' not in dir(pattern.callback):
                        continue

                    possible_methods = [m for m in dir(pattern.callback.view_class) if m in METHODS]

                    for method in possible_methods:
                        view_method_docstr = getattr(pattern.callback.view_class, method).__doc__
                        if view_method_docstr:
                            # Extract request method and JSON response from docstring of request method
                            j = re.findall(r'```(.*)```', view_method_docstr, flags=re.DOTALL)
                            if j is not None:
                                for match in j:
                                    expected_json_response = match.strip()
                                    # Only add the details if all 3 values are filled
                                    if method is not None and expected_json_response and expected_url:
                                        self.url_details.append({
                                            'url': expected_url,
                                            'method': method,
                                            'response': expected_json_response
                                        })

                except ViewDoesNotExist:
                    pass
            elif isinstance(pattern, (URLResolver, RegexURLResolver)):
                try:
                    patterns = pattern.url_patterns
                except ImportError:
                    continue

                d = describe_pattern(pattern)
                current_full_url = parent + d
                self._get_view_details(patterns, current_full_url)

    
    def _load_url_details(self):
        root_urls = self._urlconf.urls.urlpatterns
        self._get_view_details(root_urls)



import os
from setuptools import setup


here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

setup(
    name='django-rest-mock',
    version='0.1',
    packages=['rest_mock_server'],
    description='Mock Express server generated based on your views, that can come in handy when developing REST APIs with Django',
    long_description=README,
    author='Thomas Jiang',
    author_email='thomasjiangcy@gmail.com',
    url='https://github.com/thomasjiangcy/django-rest-mock-server',
    license='MIT',
    install_requires=[
        'Django>=1.11',
        'Faker>=0.8.11'
    ]
)
